﻿/**
 * @file	surface.hpp
 * @brief	サーフェス クラスの宣言およびインターフェイスの定義をします
 */

#pragma once

#include <memory.h>
#include "endianutil.hpp"
#include "font12.inl"

/**
 * @brief サーフェス クラス
 */
struct Surface
{
	int nWidth;				//!< size of the bitmap in pixels
	int nHeight;			//!< size of the bitmap in pixels
	int nBpp;				//!< bits of pixels
	int nYAlign;			//!< bytes of each line
	uint8_t* pBuffer;		//!< buffer
	uint32_t* pPalettes;	//!< palettes

	/**
	 * コンストラクタ
	 */
	Surface()
	{
		memset(this, 0, sizeof(*this));
	}

	/**
	 * コンストラクタ
	 * @param[in] parent 親
	 * @param[in] x X
	 * @param[in] y Y
	 * @param[in] width 幅
	 * @param[in] height 高さ
	 */
	Surface(const Surface* parent, int x, int y, int width, int height)
	{
		memset(this, 0, sizeof(*this));
		Bind(parent, x, y, width, height);
	}

	/**
	 * コンストラクタ
	 * @param[in] width 幅
	 * @param[in] height 高さ
	 * @param[in] bpp 色数
	 * @param[in] yalign Yアライメント
	 * @param[in] buffer バッファ
	 * @param[in] palettes パレット
	 */
	Surface(int width, int height, int bpp, int yalign, void* buffer, uint32_t* palettes = NULL)
	{
		memset(this, 0, sizeof(*this));
		Bind(width, height, bpp, yalign, buffer, palettes);
	}

	/**
	 * デストラクタ
	 */
	~Surface()
	{
		Destroy();
	}

	/**
	 * 設定
	 * @param[in] parent 親
	 * @param[in] x X
	 * @param[in] y Y
	 * @param[in] width 幅
	 * @param[in] height 高さ
	 */
	void Bind(const Surface* parent, int x, int y, int width, int height)
	{
		Destroy();

		if (parent == NULL)
		{
			return;
		}

		this->nWidth = width;
		this->nHeight = height;
		this->nBpp = parent->nBpp;
		this->nYAlign = parent->nYAlign;
		this->pBuffer = static_cast<uint8_t*>(parent->Ptr(x, y));
		this->pPalettes = parent->pPalettes;
	}

	/**
	 * バインド
	 * @param[in] width 幅
	 * @param[in] height 高さ
	 * @param[in] bpp 色数
	 * @param[in] yalign Yアライメント
	 * @param[in] buffer バッファ
	 * @param[in] palettes パレット
	 */
	void Bind(int width, int height, int bpp, int yalign, void* buffer, uint32_t* palettes)
	{
		this->nWidth = width;
		this->nHeight = height;
		this->nBpp = bpp;
		this->nYAlign = yalign;
		this->pBuffer = static_cast<uint8_t*>(buffer);
		this->pPalettes = palettes;
	}

	/**
	 * 破棄
	 */
	void Destroy()
	{
		memset(this, 0, sizeof(*this));
	}

	/**
	 * 幅
	 * @return 幅
	 */
	int Width() const
	{
		return this->nWidth;
	}

	/**
	 * 高さ
	 * @return 高さ
	 */
	int Height() const
	{
		return this->nHeight;
	}

	/**
	 * BPP
	 * @return BPP
	 */
	int Bpp() const
	{
		return this->nBpp;
	}

	/**
	 * Y アライン
	 * @return Y アライン
	 */
	int Align() const
	{
		return this->nYAlign;
	}

	/**
	 * ライン バイト数
	 * @return ライン バイト数
	 */
	unsigned int LineBytes() const
	{
		int nBpp = this->nBpp;
		if (nBpp == 15)
		{
			nBpp = 16;
		}
		return ((this->nWidth * nBpp) + 7) / 8;
	}

	/**
	 * ライン バッファ
	 * @param[in] y Y
	 * @return バッファ
	 */
	uint8_t* Line(int y) const
	{
		return this->pBuffer + (y * this->nYAlign);
	}

	/**
	 * ポインタ
	 * @param[in] x X
	 * @param[in] y Y
	 * @return ポインタ
	 */
	void* Ptr(int x, int y) const
	{
		if ((x < 0) || (y < 0) || (x >= this->nWidth) || (y >= this->nHeight))
		{
			return NULL;
		}

		int nBpp = this->nBpp;
		if (nBpp == 15)
		{
			nBpp = 16;
		}
		return this->pBuffer + ((x * nBpp) >> 3) + (y * this->nYAlign);
	}

	/**
	 * ゲット
	 * @param[in] x X
	 * @param[in] y Y
	 * @return 色
	 */
	unsigned int Get(int x, int y) const
	{
		void* ptr = Ptr(x, y);
		if (ptr == NULL)
		{
			return 0;
		}
		switch (this->nBpp)
		{
			case 1:
				{
					const uint8_t c = *(static_cast<uint8_t*>(ptr));
					return (c >> (7 - (x & 7))) & 1;
				}

			case 4:
				{
					const uint8_t c = *(static_cast<uint8_t*>(ptr));
					return (x & 1) ? (c & 15) : (c >> 4);
				}

			case 8:
				return *(static_cast<uint8_t*>(ptr));

			case 15:
			case 16:
				return *(static_cast<uint16_t*>(ptr));

			case 24:
				return *(static_cast<uint24le_t*>(ptr));

			case 32:
				return *(static_cast<uint32_t*>(ptr));
		}
		return 0;
	}

	/**
	 * 色取得
	 * @param[in] x X
	 * @param[in] y Y
	 * @return 色
	 */
	uint32_t GetColor(int x, int y) const
	{
		const unsigned int c = Get(x, y);

		uint32_t rgb = 0;
		switch (this->nBpp)
		{
			case 1:
			case 4:
			case 8:
				if (this->pPalettes)
				{
					rgb = this->pPalettes[c];
				}
				break;

			case 15:
				rgb = ((c & 0x001f) << 19) + ((c & 0x03e0) << 6) + ((c & 0x7c00) >> 7);
				rgb += (rgb >> 5) & 0x070707;
				break;

			case 16:
				rgb = ((c & 0x001f) << 19) + ((c & 0xf800) >> 8);
				rgb += (rgb >> 5) & 0x070007;
				rgb += (c & 0x07e0) << 5;
				rgb += (c & 0x0600) >> 1;
				break;

			case 24:
			case 32:
				rgb = c;
				break;
		}
		return rgb;
	}

	/**
	 * セット
	 * @param[in] x X
	 * @param[in] y Y
	 * @param[in] color 色
	 */
	void Set(int x, int y, unsigned int color)
	{
		void* ptr = Ptr(x, y);
		if (ptr == NULL)
		{
			return;
		}
		switch (this->nBpp)
		{
			case 1:
				if (color & 1)
				{
					*(static_cast<uint8_t*>(ptr)) |= (0x80 >> (x & 7));
				}
				else
				{
					*(static_cast<uint8_t*>(ptr)) &= ~(0x80 >> (x & 7));
				}
				break;

			case 4:
				if ((x & 1) == 0)
				{
					*(static_cast<uint8_t*>(ptr)) &= 0x0f;
					*(static_cast<uint8_t*>(ptr)) |= (color << 4);
				}
				else
				{
					*(static_cast<uint8_t*>(ptr)) &= 0xf0;
					*(static_cast<uint8_t*>(ptr)) |= (color & 15);
				}
				break;

			case 8:
				*(static_cast<uint8_t*>(ptr)) = static_cast<uint8_t>(color);
				break;

			case 15:
			case 16:
				*(static_cast<uint16_t*>(ptr)) = static_cast<uint16_t>(color);
				break;

			case 24:
				*(static_cast<uint24le_t*>(ptr)) = color;
				break;

			case 32:
				*(static_cast<uint32_t*>(ptr)) = color;
				break;
		}
	}

	/**
	 * クリア
	 * @param[in] color 色
	 */
	void Clear(unsigned int color)
	{
		Fill(0, 0, Width(), Height(), color);
	}

	/**
	 * Fill
	 * @param[in] x X
	 * @param[in] y Y
	 * @param[in] cx 幅
	 * @param[in] cy 高さ
	 * @param[in] color 色
	 */
	void Fill(int x, int y, int cx, int cy, unsigned int color)
	{
		for (int i = 0; i < cy; i++)
		{
			for (int j = 0; j < cx; j++)
			{
				Set(x + j, y + i, color);
			}
		}
	}

	/**
	 * Copy
	 * @param[in] x X
	 * @param[in] y Y
	 * @param[in] sx X
	 * @param[in] sy Y
	 * @param[in] cx 幅
	 * @param[in] cy 高さ
	 */
	void Copy(int x, int y, int sx, int sy, int cx, int cy)
	{
		for (int i = 0; i < cy; i++)
		{
			for (int j = 0; j < cx; j++)
			{
				Set(x + j, y + i, Get(sx + j, sy + i));
			}
		}
	}

	/**
	 * 領域コピー
	 * @param[in] surface サーフェス
	 * @param[in] px X
	 * @param[in] py Y
	 */
	void Put(const Surface& surface, int px, int py)
	{
		for (int y = 0; y < surface.Height(); y++)
		{
			for (int x = 0; x < surface.Width(); x++)
			{
				const unsigned int c = surface.Get(x, y);
				Set(x + px, y + py, c);
			}
		}
	}

	/**
	 * Set
	 * @param[in] lpPattern パターン
	 * @param[in] x X
	 * @param[in] y Y
	 * @param[in] nColor 色
	 */
	void SetPat(const uint8_t* lpPattern, int x, int y, unsigned int nColor)
	{
		if (lpPattern == NULL)
		{
			return;
		}
		const uint8_t* p = lpPattern + 2;
		const unsigned int cy = lpPattern[1];
		for (unsigned int i = 0; i < cy; i++)
		{
			uint8_t bit = 0;
			uint8_t c = 0;
			const unsigned int cx = lpPattern[0];
			for (unsigned int j = 0; j < cx; j++)
			{
				if (!bit)
				{
					bit = 0x80;
					c = *p++;
				}
				if (c & bit)
				{
					Set(x + j, y + i, nColor);
				}
				bit >>= 1;
			}
		}
	}

	/**
	 * Charactor
	 * @param[in] x X
	 * @param[in] y Y
	 * @param[in] c キャラクタ
	 * @param[in] nColor 色
	 * @param[in] bPitch ピッチ
	 * @return 幅
	 */
	int Char12(int x, int y, char c, unsigned int nColor, bool bPitch)
	{
		const uint8_t* ptr = GetFont12Ptr(c);
		if (ptr == NULL)
		{
			return 0;
		}

		int nWidth = ptr[0];

		uint8_t src[13];
		src[0] = nWidth;
		src[1] = 11;
		memcpy(src + 2, ptr + 1, 11);
		if (!bPitch)
		{
			x += (7 - nWidth) / 2;
			nWidth = 7;
		}
		SetPat(src, x, y, nColor);
		return nWidth + 1;
	}

	/**
	 * Text
	 * @param[in] x X
	 * @param[in] y Y
	 * @param[in] lpString 文字列
	 * @param[in] nColor 色
	 * @param[in] bPitch ピッチ
	 */
	void Text12(int x, int y, const char* lpString, unsigned int nColor, bool bPitch)
	{
		if (lpString == NULL)
		{
			return;
		}

		const int orgx = x;
		while (*lpString != '\0')
		{
			uint8_t c = *lpString++;
			if (c == '\n')
			{
				x = orgx;
				y += 12;
				continue;
			}

			x += Char12(x, y, c, nColor, bPitch);
		}
	}
};
