﻿/**
 * @file	sdl2window.hpp
 * @brief	SDL2 ウィンドウ クラスの宣言およびインターフェイスの定義をします
 */

#pragma once

#include <stdarg.h>
#include "misc/surface.hpp"

static const int WindowX = 640;		/*!< 幅 */
static const int WindowY = 480;		/*!< 高さ */

/**
 * @brief SDL2 ウィンドウ クラス
 */
class Sdl2Window
{
private:
	SDL_Window* m_pWindow;			/*!< SDL Window */
	SDL_Renderer* m_pRenderer;		/*!< SDL Renderer */
	SDL_Texture* m_pTexture;		/*!< SDL Texture */
	int m_px;						/*!< Text X */
	int m_py;						/*!< Text Y */

public:
	/**
	 * コンストラクタ
	 */
	Sdl2Window()
		: m_pWindow(NULL)
		, m_pRenderer(NULL)
		, m_pTexture(NULL)
		, m_px(0)
		, m_py(0)
	{
	}

	/**
	 * ウィンドウ作成
	 * @retval true 成功
	 * @retval false 失敗
	 */
	bool Create()
	{
		do
		{
			if (SDL_InitSubSystem(SDL_INIT_VIDEO) < 0)
			{
				fprintf(stderr, "Error: SDL_Init: %s\n", SDL_GetError());
				break;
			}

			m_pWindow = SDL_CreateWindow("Window", SDL_WINDOWPOS_UNDEFINED, SDL_WINDOWPOS_UNDEFINED, WindowX, WindowY, 0);
			if (m_pWindow == NULL)
			{
				break;
			}

			m_pRenderer = SDL_CreateRenderer(m_pWindow, -1, 0);
			if (m_pRenderer == NULL)
			{
				break;
			}
			SDL_RenderSetLogicalSize(m_pRenderer, WindowX, WindowY);

			m_pTexture = SDL_CreateTexture(m_pRenderer, SDL_PIXELFORMAT_RGB555, SDL_TEXTUREACCESS_STREAMING, WindowX, WindowY);
			if (m_pTexture == NULL)
			{
				break;
			}

			Clear();

			return true;
		} while (false /*CONSTCOND*/);

		Destroy();
		return false;
	}

	/**
	 * 破棄
	 */
	void Destroy()
	{
		if (m_pWindow)
		{
			SDL_DestroyWindow(m_pWindow);
			m_pWindow = NULL;
		}
		if (m_pRenderer)
		{
			SDL_DestroyRenderer(m_pRenderer);
			m_pRenderer = NULL;
		}
		if (m_pTexture)
		{
			SDL_DestroyTexture(m_pTexture);
			m_pTexture = NULL;
		}
	}

	/**
	 * クリア
	 */
	void Clear()
	{
		if (!m_pTexture)
		{
			return;
		}

		SDL_Rect rect;
		rect.x = 0;
		rect.y = 0;
		rect.w = WindowX;
		rect.h = WindowY;

		void* pPixels = NULL;
		int nPitch = 0;
		if (SDL_LockTexture(m_pTexture, &rect, &pPixels, &nPitch) == 0)
		{
			Surface s(WindowX, WindowY, 16, nPitch, pPixels);

			s.Clear(0);
			m_px = 0;
			m_py = 0;

			SDL_UnlockTexture(m_pTexture);

			Update();
		}
	}

	/**
	 * テキスト
	 * @param[in] pString 文字列
	 */
	void AddText(const char* pString)
	{
		if ((!m_pTexture) || (pString == NULL))
		{
			return;
		}

		SDL_Rect rect;
		rect.x = 0;
		rect.y = 0;
		rect.w = WindowX;
		rect.h = WindowY;

		void* pPixels = NULL;
		int nPitch = 0;
		if (SDL_LockTexture(m_pTexture, &rect, &pPixels, &nPitch) == 0)
		{
			Surface s(WindowX, WindowY, 16, nPitch, pPixels);

			do
			{
				const char c = *pString;
				if (c == '\0')
				{
					break;
				}

				pString++;
				bool lf = (c == '\n');

				const uint8_t* ptr = GetFont12Ptr(c);
				if (ptr != NULL)
				{
					const int nWidth = ptr[0];
					lf = ((m_px + nWidth) >= WindowX);
				}

				if (lf)
				{
					m_px = 0;
					m_py += 12;
					const int to = WindowY - 12;
					const int diff = m_py - to;
					if (diff > 0)
					{
						for (int y = 0; y < to; y++)
						{
							s.Copy(0, y, 0, y + diff, WindowX, 1);
						}

						s.Fill(0, to, WindowX, 12, 0);
						m_py = to;
					}
				}
				m_px += s.Char12(m_px, m_py, c, 0xffff, true);
			} while (true /*CONSTCOND*/);

			SDL_UnlockTexture(m_pTexture);

			Update();
		}
	}

	/**
	 * テキスト
	 * @param[in] pFormat フォーマット
	 */
	void AddFormat(const char* pFormat, ...)
	{
		va_list ap;
		va_start(ap, pFormat);
		char szText[4096];
#if (_MSC_VER >= 1500 /*Visual Studio 2008*/)
		vsprintf_s(szText, pFormat, ap);
#else	// (_MSC_VER >= 1500 /*Visual Studio 2008*/)
		vsprintf(szText, pFormat, ap);
#endif	// (_MSC_VER >= 1500 /*Visual Studio 2008*/)
		va_end(ap);
		AddText(szText);
	}

	/**
	 * 更新
	 */
	void Update()
	{
		if (m_pRenderer)
		{
			SDL_RenderClear(m_pRenderer);
			if (m_pTexture)
			{
				SDL_Rect rect;
				rect.x = 0;
				rect.y = 0;
				rect.w = WindowX;
				rect.h = WindowY;
				SDL_RenderCopy(m_pRenderer, m_pTexture, &rect, NULL);
			}
			SDL_RenderPresent(m_pRenderer);
		}
	}
};
